#/include/rt.g

/*
 * fltio.d: Draco run-time routines for floating point I/O.
 */

extern
    _d_channelPutChar(char ch)void,
    _d_charsput(*char string)void,
    _d_channelGetChar()char,
    _d_channelUnGetChar(char ch)void,
    _d_channelError(CHANNELERROR ecode)void,
    _d_channelSkip()void;

uint
    FLOAT_BYTES = 8,
    FLOAT_DIGITS = 17,
    DEFAULT_F_DIGITS = 5;

extern
    _d_bufferToFloat(*char buffer; int decExp; *float pF)uint,
    _d_floatToBuffer(float f; *char pBuff)int;

extern
    _d_fltFormat(float f; int len, digits; char format)void;

/*
 * _d_fltfgput - G format - use whichever looks best.
 */

proc _d_fltfgput(float f; int len, digits)void:

    _d_fltFormat(f, len, digits, 'g');
corp;

/*
 * _d_fltffput - F format.
 */

proc _d_fltffput(float f; int len, digits)void:

    _d_fltFormat(f, len, digits, 'f');
corp;

/*
 * _d_fltfeput - E format.
 */

proc _d_fltfeput(float f; int len, digits)void:

    _d_fltFormat(f, len, digits, 'e');
corp;

/*
 * _d_fltFormat - common formatting routine for all formats.
 */

proc _d_fltFormat(float f; register int len, digits; char format)void:
    extern
	_d_floatRound(*[FLOAT_DIGITS] char buff; uint sig)bool;
    uint
	E_OVERHEAD = 8, 		/* +0.e+ddd */
	F_OVERHEAD = 1; 		/* just the decimal point */
    register proc(char ch)void r_d_channelPutChar;
    [FLOAT_BYTES] byte floatBuff @ f;
    [FLOAT_BYTES / sizeof(ulong)] ulong longBuff @ f;
    [FLOAT_DIGITS] char buff;
    register *char p;
    register int sig, i;	/* i is int, so that 'for' can not do */
    int exp, overHead;
    bool isNeg, pad1, pad2, hadDot, hadFraction;

    r_d_channelPutChar := _d_channelPutChar;

    /* get rid of any sign */

    if floatBuff[0] & 0x80 ~= 0 then
	floatBuff[0] := floatBuff[0] & 0x7f;
	isNeg := true;
    else
	isNeg := false;
    fi;

    /* extract padding information from formatting parameters */

    if len < 0 then
	len := - len;
	pad1 := true;
    else
	pad1 := false;
    fi;
    if digits < 0 then
	digits := - digits;
	pad2 := true;
    else
	pad2 := false;
    fi;

    /* check for Inf's and NaN's */

    if floatBuff[0] = 0x7f and floatBuff[1] & 0xf0 = 0xf0 then
	if len ~= 0 and len < 4 then
	    for i from 1 upto len do
		r_d_channelPutChar('*');
	    od;
	else
	    for i from 1 upto len - 4 do
		r_d_channelPutChar(' ');
	    od;
	    if isNeg then
		r_d_channelPutChar('-');
	    else
		r_d_channelPutChar('+');
	    fi;
	    if FLOAT_BYTES ~= sizeof(ulong) * 2 then
		error("Quick check for Inf won't work");
	    fi;
	    if longBuff[0] & 0x000fffff = 0 and longBuff[1] = 0 then
		/* it is an Inf */
		r_d_channelPutChar('I');
		r_d_channelPutChar('n');
		r_d_channelPutChar('f');
	    else
		/* is a NaN */
		r_d_channelPutChar('N');
		r_d_channelPutChar('a');
		r_d_channelPutChar('N');
	    fi;
	fi;
    else

	/* get the decimal digits and exponent */

	exp := _d_floatToBuffer(f, &buff[0]);
	if buff[0] = '0' and (longBuff[0] ~= 0 or longBuff[1] ~= 0) then
	    /* a leading 0 is present - shift the whole thing left 1 digit */
	    exp := exp - 1;
	    p := &buff[1];
	    for i from FLOAT_DIGITS - 2 downto 0 do
		(p - sizeof(char))* := p*;
		p := p + sizeof(char);
	    od;
	    (p - sizeof(char))* := '0';
	fi;

	/* determine which format we are going to use */

	if format = 'e' or format = 'g' and (exp < -5 or exp > 10) then

	    /* use E format */

	    if digits = 0 then
		if len = 0 then
		    sig := FLOAT_DIGITS - 2;
		else
		    sig := len - E_OVERHEAD;
		fi;
	    else
		sig := digits;
	    fi;
	    if sig > FLOAT_DIGITS then
		sig := FLOAT_DIGITS;
	    elif sig < 1 then
		sig := 1;
	    fi;

	    /* we have the needed number of digits - round to that */

	    if sig ~= FLOAT_DIGITS and buff[sig] >= '5' and
		_d_floatRound(&buff, sig)
	    then
		exp := exp + 1;
	    fi;

	    /* for free-format, get rid of trailing 0's */

	    if len = 0 and digits = 0 then
		while sig ~= 1 and buff[sig - 1] = '0' do
		    sig := sig - 1;
		od;
	    fi;

	    /* final output of E format value */

	    for i from len - sig - (E_OVERHEAD + 1) downto 0 do
		r_d_channelPutChar(' ');
	    od;
	    _d_channelPutChar(
		if isNeg then
		    '-'
		elif pad1 then
		    '+'
		else
		    ' '
		fi);
	    r_d_channelPutChar('0');
	    r_d_channelPutChar('.');
	    for i from 0 upto sig - 1 do
		r_d_channelPutChar(buff[i]);
	    od;
	    r_d_channelPutChar('e');
	    if exp < 0 then
		exp := -exp;
		r_d_channelPutChar('-');
	    else
		r_d_channelPutChar('+');
	    fi;
	    r_d_channelPutChar(exp / 100 + '0');
	    r_d_channelPutChar(exp / 10 % 10 + '0');
	    r_d_channelPutChar(exp % 10 + '0');

	else

	    /* use F format */

	    overHead := if isNeg then F_OVERHEAD + 1 else F_OVERHEAD fi;

	    if len = 0 then

		/* free format - use minimum reasonable length */

		if digits = 0 then
		    if exp > DEFAULT_F_DIGITS then
			sig := exp;
		    else
			sig := DEFAULT_F_DIGITS;
		    fi;
		else
		    sig := exp + digits;	/* works for exp + and - */
		fi;
		if sig > FLOAT_DIGITS then
		    sig := FLOAT_DIGITS;
		elif sig < 1 then
		    sig := 1;
		fi;
		/* round the value */
		if sig ~= FLOAT_DIGITS and buff[sig] >= '5' and
		    _d_floatRound(&buff, sig)
		then
		    exp := exp + 1;
		fi;
		/* trim trailing zeros */
		while sig ~= 1 and buff[sig - 1] = '0' do
		    sig := sig - 1;
		od;
		if isNeg then
		    r_d_channelPutChar('-');
		fi;
		if exp <= 0 then
		    r_d_channelPutChar('0');
		    r_d_channelPutChar('.');
		    while exp < 0 do
			exp := exp + 1;
			digits := digits - 1;
			r_d_channelPutChar('0');
		    od;
		    for i from 0 upto sig - 1 do
			r_d_channelPutChar(buff[i]);
			digits := digits - 1;
		    od;
		    while digits > 0 do
			digits := digits - 1;
			if pad2 then
			    r_d_channelPutChar('0');
			else
			    r_d_channelPutChar(' ');
			fi;
		    od;
		else
		    hadDot := false;
		    for i from 0 upto sig - 1 do
			r_d_channelPutChar(buff[i]);
			exp := exp - 1;
			if exp = 0 then
			    r_d_channelPutChar('.');
			    hadDot := true;
			    if digits = 0 and i = sig - 1 then
				r_d_channelPutChar('0');
			    fi;
			fi;
			if hadDot then
			    digits := digits - 1;
			fi;
		    od;
		    if not hadDot then
			while exp > 0 do
			    exp := exp - 1;
			    r_d_channelPutChar('0');
			od;
			r_d_channelPutChar('.');
			r_d_channelPutChar('0');
		    fi;
		    while digits > 0 do
			digits := digits - 1;
			if pad2 then
			    r_d_channelPutChar('0');
			else
			    r_d_channelPutChar(' ');
			fi;
		    od;
		fi;

	    else

		/* we are given specific len and digits to follow */

		sig := exp + digits;	    /* works for + and - exp */
		if sig > FLOAT_DIGITS then
		    sig := FLOAT_DIGITS;
		elif sig <= 0 then
		    sig := 1;
		    /* in the missing else below, we will, below, be
		       rounding in such a way that 'exp' will be
		       incremented, which will mean that we wanted one
		       more digit than otherwise */
		    if buff[0] ~= '9' or buff[1] < '5' then
			buff[0] := '0';
			buff[1] := '0';
			exp := 0;
			sig := 0;
		    fi;
		fi;
		/* round down to required sig digits */
		if sig ~= 0 and sig ~= FLOAT_DIGITS and
		    buff[sig] >= '5' and _d_floatRound(&buff, sig)
		then
		    exp := exp + 1;
		    if exp + digits <= 0 then
			/* we really do want to end up with a 0 */
			buff[0] := '0';
			buff[1] := '0';
			exp := 0;
			sig := 0;
		    fi;
		fi;

		/* we delay this test until here, since 'exp' can be changed
		   above. */

		if exp >= 0 and len < exp + digits + overHead or
		    exp < 0 and len < digits + overHead
		then

		    /* value will not fit */

		    for i from 1 upto len do
			r_d_channelPutChar('*');
		    od;

		else

		    /* trim trailing zeros */

		    while sig > 1 and buff[sig - 1] = '0' do
			sig := sig - 1;
		    od;

		    /* final print of formatted value */

		    hadDot := false;
		    hadFraction := false;
		    if pad1 then
			if isNeg then
			    r_d_channelPutChar('-');
			fi;
			if exp <= 0 then
			    for i from 1 upto len - overHead - digits do
				r_d_channelPutChar('0');
			    od;
			    r_d_channelPutChar('.');
			    hadDot := true;
			    while exp < 0 do
				exp := exp + 1;
				r_d_channelPutChar('0');
				digits := digits - 1;
				hadFraction := true;
			    od;
			else
			    for i from 1 upto len - overHead - digits - exp do
				r_d_channelPutChar('0');
			    od;
			fi;
		    else
			if exp <= 0 then
			    for i from 1 upto len - overHead - digits - 1 do
				r_d_channelPutChar(' ');
			    od;
			    if isNeg then
				r_d_channelPutChar('-');
			    fi;
			    if len - overHead - digits > 0 then
				r_d_channelPutChar('0');
			    fi;
			    r_d_channelPutChar('.');
			    hadDot := true;
			    while exp < 0 do
				exp := exp + 1;
				r_d_channelPutChar('0');
				digits := digits - 1;
				hadFraction := true;
			    od;
			else
			    for i from 1 upto len - overHead - digits - exp do
				r_d_channelPutChar(' ');
			    od;
			    if isNeg then
				r_d_channelPutChar('-');
			    fi;
			fi;
		    fi;

		    for i from 0 upto sig - 1 do
			r_d_channelPutChar(buff[i]);
			if hadDot then
			    hadFraction := true;
			    digits := digits - 1;
			fi;
			exp := exp - 1;
			if exp = 0 then
			    r_d_channelPutChar('.');
			    hadDot := true;
			fi;
		    od;
		    if not hadDot then
			while exp > 0 do
			    exp := exp - 1;
			    r_d_channelPutChar('0');
			od;
			r_d_channelPutChar('.');
		    fi;
		    while digits > 0 do
			digits := digits - 1;
			if pad2 or not hadFraction then
			    hadFraction := true;
			    r_d_channelPutChar('0');
			else
			    r_d_channelPutChar(' ');
			fi;
		    od;
		fi;
	    fi;
	fi;
    fi;
corp;

/*
 * _d_floatRound - used by _d_fltput to round to a given number of digits.
 *	Return 'true' if we need to increment the exponent.
 */

proc _d_floatRound(*[FLOAT_DIGITS] char buff; uint sig)bool:
    register *char p;
    register uint i;

    i := sig;
    p := &buff*[i - 1];
    while i ~= 0 and p* = '9' do
	p* := '0';
	p := p - sizeof(char);
	i := i - 1;
    od;
    if i = 0 then
	/* we have gone all the way to the upper digit and turned it
	   into a '0'. We shift right so we can insert a '1'. */
	p := &buff*[FLOAT_DIGITS];
	for i from FLOAT_DIGITS - 2 downto 0 do
	    p := p - sizeof(char);
	    p* := (p - sizeof(char)) *;
	od;
	(p - sizeof(char))* := '1';
	true
    else
	/* just increment the current digit */
	p* := p* + 1;
	false
    fi
corp;

/*
 * _d_fltget - read a float.
 */

proc _d_fltget(*float pF)void:
    register proc()char r_d_channelGetChar;
    float f;
    register uint len;
    register int exp, expAdjust;
    [FLOAT_BYTES] byte floatBuff @ f;
    [FLOAT_DIGITS + 1] char buff;
    register char ch;
    byte digit;
    CHANNELERROR err;
    bool hadZero, sig, dot, expNeg, isNegative;

    r_d_channelGetChar := _d_channelGetChar;
    err := CH_OK;
    _d_channelSkip();
    ch := r_d_channelGetChar();
    if ch = '\e' then
	err := CH_EOF;
    elif ch = '\n' then
	err := CH_MISSING;
    fi;
    isNegative := false;
    if ch = '-' then
	ch := r_d_channelGetChar();
	isNegative := true;
    elif ch = '+' then
	ch := r_d_channelGetChar();
    fi;
    hadZero := false;
    if ch = '0' then
	hadZero := true;
	ch := r_d_channelGetChar();
	if ch = 'f' or ch = 'F' then
	    ch := r_d_channelGetChar();
	fi;
    fi;
    if ch = 'x' or ch = 'X' then
	/* hex form, must be FLOAT_BYTES * 2 hex digits */
	len := 0;
	ch := r_d_channelGetChar();
	while ch >= '0' and ch <= '9' or ch >= 'A' and ch <= 'F' or
	    ch >= 'a' and ch <= 'f'
	do
	    digit :=
		if ch >= 'a' then
		    ch - ('a' - 10)
		elif ch >= 'A' then
		    ch - ('A' - 10)
		else
		    ch - '0'
		fi;
	    if len < FLOAT_BYTES * 2 then
		if len % 2 = 0 then
		    floatBuff[len / 2] := digit;
		else
		    floatBuff[len / 2] := floatBuff[len / 2] << 4 | digit;
		fi;
		len := len + 1;
	    else
		err := CH_OVERFLOW;
	    fi;
	    ch := r_d_channelGetChar();
	od;
	if len ~= FLOAT_BYTES * 2 then
	    err := CH_BADCHAR;
	fi;
    elif ch = 'N' or ch = 'n' then
	ch := r_d_channelGetChar();
	if ch = 'a' or ch = 'A' then
	    ch := r_d_channelGetChar();
	    if ch = 'N' or ch = 'n' then
		ch := r_d_channelGetChar();
		for len from 1 upto FLOAT_BYTES - 1 do
		    floatBuff[len] := 0xff;
		od;
		floatBuff[0] := 0x7f;
	    else
		err := CH_BADCHAR;
	    fi;
	else
	    err := CH_BADCHAR;
	fi;
    elif ch = 'I' or ch = 'i' then
	ch := r_d_channelGetChar();
	if ch = 'n' or ch = 'N' then
	    ch := r_d_channelGetChar();
	    if ch = 'f' or ch = 'F' then
		ch := r_d_channelGetChar();
		for len from 2 upto FLOAT_BYTES - 1 do
		    floatBuff[len] := 0;
		od;
		floatBuff[0] := 0x7f;
		floatBuff[1] := 0xf0;
	    else
		err := CH_BADCHAR;
	    fi;
	else
	    err := CH_BADCHAR;
	fi;
    elif ch = '.' or ch >= '0' and ch <= '9' or hadZero then

	/* set up for and get the decimal digits */

	for len from 0 upto FLOAT_DIGITS do
	    buff[len] := '0';
	od;
	len := 0;
	expAdjust := - (FLOAT_DIGITS + 1);
	sig := false;
	dot := false;
	while
	    if ch = '.' and not dot then
		dot := true;
		ch := r_d_channelGetChar();
	    fi;
	    ch >= '0' and ch <= '9'
	do
	    if ch ~= '0' then
		sig := true;
	    fi;
	    if sig and len ~= FLOAT_DIGITS + 1 then
		buff[len] := ch;
		len := len + 1;
	    fi;
	    if not dot and sig then
		expAdjust := expAdjust + 1;
	    elif dot and not sig then
		expAdjust := expAdjust - 1;
	    fi;
	    ch := r_d_channelGetChar();
	od;

	/* get any decimal exponent given */

	exp := 0;
	if ch = 'e' or ch = 'E' then
	    ch := r_d_channelGetChar();
	    expNeg := false;
	    if ch = '+' then
		ch := r_d_channelGetChar();
	    elif ch = '-' then
		ch := r_d_channelGetChar();
		expNeg := true;
	    fi;
	    if ch < '0' or ch > '9' then
		err := CH_BADCHAR;
	    fi;
	    while ch >= '0' and ch <= '9' do
		if exp < 40 then
		    exp := exp * 10 + (ch - '0');
		else
		    err := CH_OVERFLOW;
		fi;
		ch := r_d_channelGetChar();
	    od;
	    if expNeg then
		exp := -exp;
	    fi;
	fi;
	exp := exp + expAdjust;
	if exp > 309 then
	    err := CH_OVERFLOW;
	elif exp < -323 then
	    err := CH_OVERFLOW;
	fi;

	/* attempt to get a final decimal exponent of 0, so that we can
	   avoid a floating point multiply (special cased in IEEE lib) */

	while exp < 0 and buff[FLOAT_DIGITS] = '0' do
	    exp := exp + 1;
	    for len from FLOAT_DIGITS downto 1 do
		buff[len] := buff[len - 1];
	    od;
	    buff[0] := '0';
	od;

	/* convert to IEEE 64 bit floating point */

	if _d_bufferToFloat(&buff[0], exp, &f) ~= 0 then
	    err := CH_OVERFLOW;
	fi;
    elif err = CH_OK then
	err := CH_BADCHAR;
    fi;
    if isNegative then
	floatBuff[0] := floatBuff[0] | 0x80;
    fi;
    if err ~= CH_OK then
	_d_channelError(err);
    fi;
    _d_channelUnGetChar(ch);
    pF* := f;
corp;
