	xdef	_d_floatToBuffer
	xref	_MathIeeeDoubBasBase
	xref	_d_flt_powTab

_LVOIEEEDPMul	equ	-78

* _d_floatToBuffer
*	The first parameter is a positive floating point number.
*	The second (Draco calling convention) is the address of a 17 byte
*	buffer to put the decimal digits into. Returned is a 16 bit signed
*	value which is the decimal exponent, given a decimal point to the
*	left of the 17 digits.

_d_floatToBuffer
	movem.l d2/d3/d4/d5/a6,-(sp)
	movem.l 28(sp),d0/d1		; fetch the floating point value
	move.l	d0,d2
	or.l	d1,d2			; is value an exact 0?
	bne.s	notZero 		; no => go process normally
	moveq.l #0,d5			; use 0 as decimal exponent
	move.l	24(sp),a0		; address of result buffer
	bra.l	noShift 		; and skip most of the work
notZero
	move.l	_MathIeeeDoubBasBase,a6 ; ready to call IEEEDP library
	move.l	d0,d5
	swap	d5
	andi.w	#$7ff0,d5		; isolate exponent
	lsl.w	#1,d5			; put into high 11 bits
	subi.w	#1023*32,d5		; adjust for bias
	asr.w	#5,d5			; make into signed 16 bit value
	muls	#19728,d5		; mul by ln 2 x 2**16
	swap	d5			; get proper bits
	move.w	d5,d4
	asl.w	#3,d4			; * 8 (8 bytes per value)
	lea	_d_flt_powTab,a0	; address of 10**0
expLoop
	movem.l 0(a0,d4.w),d2/d3	; load power of 10
	cmp.l	d0,d2			; can use a simple compare -
	bgt.s	expOK			; don't need to use floating point
	blt.s	expInc
	cmp.l	d1,d3			; high parts equal, test low parts
	bgt.s	expOK
expInc	addq.w	#1,d5
	addq.w	#8,d4			; bump exponent
	cmpi.w	#308,d5 		; exponent too large?
	blt.s	expLoop 		; no => keep looping
expOK
	neg.w	d4
	movem.l 0(a0,d4.w),d2/d3	; get 10 ** (-exp + 17)
	jsr	_LVOIEEEDPMul(a6)	; get final scaled value

* Turn the binary value into a decimal one. We remove the exponent, and add
* in the implicit bit (the one not stored in IEEE format). Then, we shift
* the fraction left by 0, 1, 2, or 3 bits, so that the exponent is always
* the same (for a given decimal exponent, the binary exponent can be any
* of 4 different values). We go through the full 64 bits, doing multiplies by
* 10. '0' is added to the integer part to produce the decimal digits which
* are stored into the output buffer.
* After the shift the binary point is between the two high order bytes.
* Therefore the decimal digits appear in the high order byte.

	move.l	24(sp),a0		; address of result buffer
	move.l	d0,d2
	or.l	d1,d2			; any bits set?
	beq.s	noShift 		; found a real zero.  don't shift it
	move.l	d0,d2
	andi.l	#$000fffff,d0		; remove exponent
	ori.l	#$00100000,d0		; add in implicit bit
	swap	d2
	andi.w	#$7ff0,d2		; isolate exponent
	subi.w	#$3fb0,d2		; move binary point
	lsr.w	#4,d2			; to between the high
	bra.s	test			; order
shiftLoop				; bytes
	lsl.l	#1,d1			; (Except for values too big!)
	roxl.l	#1,d0
test
	dbra	d2,shiftLoop

* produce decimal digits, and store in output buffer

	move.l	d0,d2
	moveq.l #24,d3
	lsr.l	d3,d2
	beq.s	noShift 		; no => normal processing
	moveq.l #17-2,d4		; one less digits than normal
*	divu.w	#10,d2			; produce quotient, remainder
*	beq.s	oneExtra		; quotient 0, only 1 extra digit
	addi.b	#'0',d2
	move.b	d2,(a0)+		; store first digit
*	moveq.l #17-3,d4		; two less digits than normal
	addq.w	#1,d5			; bump decimal exponent
*oneExtra
*	swap	d2			; remainder to low word
*	addi.b	#'0',d2
*	move.b	d2,(a0)+		; store second digit
*	addq.w	#1,d5			; bump decimal exponent
	andi.l	#$00ffffff,d0		; remove the stuff we've used
	bra.s	digLoop 		; don't reload loop count
noShift
	moveq.l #17-1,d4		; loop count: 17 digits
digLoop
	lsl.l	#1,d1
	roxl.l	#1,d0
	move.l	d0,d2
	move.l	d1,d3
	lsl.l	#1,d1
	roxl.l	#1,d0
	lsl.l	#1,d1
	roxl.l	#1,d0
	add.l	d3,d1
	addx.l	d2,d0
	move.l	d0,d2
	andi.l	#$00ffffff,d0
	swap	d2
	lsr.w	#8,d2
	addi.b	#'0',d2
	move.b	d2,(a0)+
	dbra	d4,digLoop		; repeat for 17 decimal digits

	move.w	d5,d0			; return decimal exponent
	movem.l (sp)+,d2/d3/d4/d5/a6	; restore regs
	move.l	(sp)+,a0		; return address
	adda.l	#12,sp			; pop parameters
	jmp	(a0)			; return

	end
