	xdef	_d_bufferToFloat
	xref	_MathIeeeDoubBasBase
	xref	_d_flt_t308,_d_flt_tm308,_d_flt_powTab

_LVOIEEEDPMul	equ	-78

* _d_bufferToFloat
*	the 18 digits in the first parameter (Draco calling convention)
*	are the decimal form of the input floating point value. The decimal
*	point is considered to be to the right of them. The second parameter
*	is the signed 16 bit binary of the decimal exponent to be added in.
*	The third and last parameter is the place to store the 8 bytes of
*	the final floating point value. The result returned in D0 is a
*	16 bit value: 0 - all OK, 1 - overflow, 2 - underflow.

_d_bufferToFloat
	movem.l d2/d3/d4/d5/d6/a6,-(sp)

* first, convert the 18 digits into a 64 bit binary value in d0/d1

	move.l	34(sp),a0		; address of digits
	clr.l	d0			; clear "accumulator"
	clr.l	d1
	clr.l	d5
	clr.l	d6
	moveq.l #18-1,d4		; there will be 18 digits
digLoop
	lsl.l	#1,d1
	roxl.l	#1,d0			; shift left 1 bit (multiply by 2)
	move.l	d0,d2
	move.l	d1,d3			; make a copy in d2/d3
	lsl.l	#1,d3
	roxl.l	#1,d2			; shift that left 1 (now is orig * 4)
	lsl.l	#1,d3
	roxl.l	#1,d2			; and one more (now is original * 8)
	add.l	d3,d1
	addx.l	d2,d0			; add to * 2 set => yields mul by 10
	move.b	(a0)+,d5		; fetch digit
	sub.b	#'0',d5 		; turn into straight binary
	add.l	d5,d1
	addx.l	d6,d0			; add in digit
	dbf	d4,digLoop		; repeat for 18 digits

* make sure the high 11 bits are clear for the exponent insertion

	move.l	d0,d2
	or.l	d1,d2
	beq.l	exit			; value is 0 => quick exit
	move.l	#$3fe+53,d4		; default exponent (53 bits here)
fixExp
	clr.b	d5			; clear rounding bit flag
expLoop
	move.l	d0,d2
	and.l	#$FFE00000,d2
	beq.s	expOK			; no need to trim anything
	addq.l	#1,d4			; bump exponent
	lsr.l	#1,d0
	roxr.l	#1,d1			; shift fraction right
	scs	d5			; set rounding flag according to carry
	bra.s	expLoop 		; repeat as needed
expOK
	tst.b	d5			; do we need to round up?
	beq.s	noRoundUp		; branch if we don't need to
	addq.l	#2,d1			; have to add to 64 bit value
	addx.l	d6,d0
	bra.s	fixExp			; may have overflowed 53 bits
					; rounding cannot occur again, so
					; will not get incorrect result.
noRoundUp

* make sure we get a single 1 bit in the proper position

fracLoop
	btst.l	#20,d0
	bne.s	fracOK
	subq.l	#1,d4			; decrement exponent
	lsl.l	#1,d1
	roxl.l	#1,d0			; shift fraction left one bit
	bra.s	fracLoop		; repeat as needed

fracOK
	moveq.l #20,d5
	lsl.l	d5,d4			; shift exponent to proper position
	and.l	#$000FFFFF,d0		; clear the implied 1-bit
	or.l	d4,d0			; OR in the base exponent

* set up to call the IEEE double precision multiply routine

	movea.l _MathIeeeDoubBasBase,a6

* now fetch the supplied exponent, fix it up if needed, and apply it

	move.w	32(sp),d4		; argument exponent value

shrinkExp
	cmp.w	#308,d4 		; is it too big?
	blt	notTooBig
	sub.w	#308,d4 		; shrink exponent, and
	movem.l _d_flt_t308,d2/d3	; multiply by 1.0E+308
	jsr	_LVOIEEEDPMul(a6)
	bra.s	shrinkExp		; repeat as needed
notTooBig

expandExp
	cmp.w	#-308,d4		; is it too small?
	bgt	notTooSmall
	add.w	#308,d4 		; bump exponent, and
	movem.l _d_flt_tm308,d2/d3	; multiply by 1.0E-308
	jsr	_LVOIEEEDPMul(a6)
	bra.s	expandExp
notTooSmall

	asl.w	#3,d4			; * 8 (8 bytes per table value)
	lea.l	_d_flt_powTab,a0
	movem.l 0(a0,d4.w),d2/d3	; multiply by 10 to the exponent power
	jsr	_LVOIEEEDPMul(a6)

* all done - store result back and exit

exit
	move.l	28(sp),a0		; where to store result
	movem.l d0/d1,0(a0)		; store it
	movem.l (sp)+,d2/d3/d4/d5/d6/a6 ; restore regs
	move.l	(sp)+,a0		; return address
	adda.w	#10,sp			; pop parameters
	clr.w	d0			; no error (can't seem to detect one!)
	jmp	0(a0)			; return

	end
