/*
 * RT.G - include file for Draco run-time system.
 */

type

    FileHandle_t = ulong,

    /* structure of Draco's 'file' type on AmigaDOS systems: */

    FILE = struct {
	ulong f_buffPos;	/* position for next I/O operation */
	ulong f_inMax;		/* amount of data now in the buffer */
	ulong f_outMax; 	/* last written position in buffer */
	ulong f_buffOffset;	/* which byte is at start of buffer */
	ulong f_buffLen;	/* length of buffer associated with file */
	*char f_name;		/* pointer to saved file name */
	FileHandle_t f_fd;	/* the AmigaDOS fd for this file */
	[2] char f_buffer;	/* actually however big user makes it */
    },

    /* error codes returned by 'IOerror': */

    CHANNELERROR = enum {
	CH_OK,			/* no error */

	CH_EOF, 		/* read past end-of-file indicator */
	CH_CLOSED,		/* use after close */

	CH_NONEXIS,		/* file doesn't exist */
	CH_DISKFULL,		/* disk is full; write failed */
	CH_BADSEEK,		/* a seek call failed */

	CH_MISSING,		/* no data on line */
	CH_BADCHAR,		/* bad character for input conversion */
	CH_OVERFLOW,		/* overflow on numeric conversion */
	CH_UNDERFLOW,		/* not used - floating point input */

	CH_BADREAD,		/* the read call failed! */
    },

    /* the united type of the units possible for a channel: */

    CHANNELUNIT = union {
	*FILE ch_file;
	proc()byte ch_getByte;
	proc()char ch_getChar;
	proc(byte b)void ch_putByte;
	proc(char ch)void ch_putChar;
	*char ch_chars;
    },

    /* structure of Draco's 'channel' type on AmigaDOS systems: */

    CHANNEL = struct {
	CHANNELUNIT ch_;
	enum {
	    CH_STANDARD,	/* text I/O to or from std input (console) */
	    CH_FILE,		/* text or binary I/O to or from a file */
	    CH_RANDOMFILE,	/* binary I/O to and from a file */
	    CH_PROC,		/* text or binary I/O through a procedure */
	    CH_CHARS		/* text I/O to or from a "chars" value */
	} ch_mode;
	bool ch_input;		/* true => input chanel */
	bool ch_text;		/* true => text channel */
	bool ch_eof;		/* true => eof reached on this channel */
	bool ch_closed; 	/* true => channel has been closed */
	CHANNELERROR ch_error, ch_newError;  /* error before, error new */
	bool ch_gotChar;	/* true => got one char readahead */
	char ch_buffer; 	/* buffer for one character readahead */
    };

/* magic constant returned by LineRead on EOF: */

ulong LINE_EOF = 0xffffffff;

/* A magic number I use to say that we haven't done anything yet: */

long POS_UNKNOWN       = 0Lxffffffff;
