#/include/rt.g
#drinc:libraries/dos.g

/* Parameterize the stuff we do here: */

uint
    CHANNEL_STACK_SIZE = 10,		/* max nesting of read/write */
    LINE_BUFFER_SIZE   = 256;		/* max length of terminal line */

/* Our local variables: */

[CHANNEL_STACK_SIZE] *CHANNEL ChannelStack;

unsigned CHANNEL_STACK_SIZE ChannelStackNext;

CHANNEL
    StandardInputChannel,
    StandardOutputChannel;

*CHANNEL
    StandardInput,
    StandardOutput;

FileHandle_t
    StandardInputFd,
    StandardOutputFd;

[LINE_BUFFER_SIZE] char LineInputBuffer, LineOutputBuffer;
unsigned LINE_BUFFER_SIZE
    LineInputMax,
    LineInputPos,
    LineOutputPos;

ulong
    NormalWriteOpenMode, NormalReadOpenMode,
    PipeWriteOpenMode, PipeReadOpenMode;

/*
 * _d_setstdin - set current channel to be the standard input.
 */

proc _d_setstdin()void:
    extern
	LineFlush()void,
	_d_setchannel(*CHANNEL chan)void;

    LineFlush();
    _d_setchannel(StandardInput);
corp;

/*
 * _d_setstdout - set current channel to be the standard output.
 */

proc _d_setstdout()void:
    extern
	_d_setchannel(*CHANNEL chan)void;

    _d_setchannel(StandardOutput);
corp;

/*
 * _d_setchannel - use the passed CHANNEL as the current I/O channel.
 */

proc _d_setchannel(*CHANNEL chan)void:
    extern
	_d_printString(*char mess)void;

    if ChannelStackNext = CHANNEL_STACK_SIZE then
	_d_printString("\n****Draco: read/writes nested too deep.****\n");
	Exit(0L35);
    fi;
    ChannelStack[ChannelStackNext] := chan;
    ChannelStackNext := ChannelStackNext + 1;
corp;

/*
 * _d_unchannel - close off I/O on the current channel. Return true if no
 *		  error flag has been raised.
 */

proc _d_unchannel()bool:
    register *CHANNEL chan;

    ChannelStackNext := ChannelStackNext - 1;
    chan := ChannelStack[ChannelStackNext];
    chan*.ch_error := chan*.ch_newError;
    chan*.ch_newError := CH_OK;
    chan*.ch_error = CH_OK
corp;

/*
 * _d_channelError - set an error code into a channel.
 */

proc _d_channelError(CHANNELERROR eCode)void:
    register *CHANNEL chan;

    chan := ChannelStack[ChannelStackNext - 1];
    if chan*.ch_newError = CH_OK then
	chan*.ch_newError := eCode;
    fi;
corp;

/*
 * _d_channelHadError - return 'true' if channel has had an error recently.
 */

proc _d_channelHadError()bool:

    ChannelStack[ChannelStackNext - 1]*.ch_newError ~= CH_OK
corp;

/*
 * _d_stdIOerror - return the error code associated with the standard input.
 */

proc _d_stdIOerror()ushort:
    CHANNELERROR eCode;
    register *CHANNEL chan;

    chan := StandardInput;
    eCode := chan*.ch_error;
    chan*.ch_error := CH_OK;
    if chan*.ch_eof and chan*.ch_gotChar and chan*.ch_buffer = '\e' then
	chan*.ch_gotChar := false;
    fi;
    chan*.ch_eof := false;
    eCode - CH_OK
corp;

/*
 * _d_channelGetChar - get the next character from the current text channel.
 */

proc _d_channelGetChar()char:
    extern _d_readErr()void;
    register *CHANNEL chan;
    CHANNELUNIT unit;
    register ulong pos;
    register long len;
    register char ch;

    chan := ChannelStack[ChannelStackNext - 1];
    if chan*.ch_error ~= CH_OK then
	_d_readErr();
	'\(0xff)'
    elif chan*.ch_newError ~= CH_OK then
	'\(0xff)'
    elif chan*.ch_closed then
	chan*.ch_newError := CH_CLOSED;
	'\(0xff)'
    elif chan*.ch_gotChar then
	chan*.ch_gotChar := false;
	chan*.ch_buffer
    elif chan*.ch_eof then
	chan*.ch_newError := CH_EOF;
	'\e'
    else
	/* no special case, have to actually get a character. */
	unit.ch_chars := chan*.ch_.ch_chars;
	case chan*.ch_mode
	incase CH_STANDARD:
	    if LineInputPos = LineInputMax then
		LineInputPos := 0;
		len :=
		    Read(StandardInputFd, &LineInputBuffer[0],
			 LINE_BUFFER_SIZE);
		if len = -0L1 then
		    LineInputMax := 0;
		    ch := '\(0xff)';
		    chan*.ch_newError := CH_BADREAD;
		else
		    LineInputMax := len;
		    if LineInputMax = 0 then
			ch := '\e';
		    else
			ch := LineInputBuffer[0];
			LineInputPos := 1;
		    fi;
		fi;
	    else
		LineInputPos := LineInputPos + 1;
		ch := LineInputBuffer[LineInputPos - 1];
	    fi;
	incase CH_FILE:
	    pos := unit.ch_file*.f_buffPos;
	    if pos = unit.ch_file*.f_inMax then
		unit.ch_file*.f_buffPos := 0L0;
		len := Read(unit.ch_file*.f_fd, &unit.ch_file*.f_buffer[0],
			    unit.ch_file*.f_buffLen);
		if len = -0L1 then
		    unit.ch_file*.f_inMax := 0;
		    ch := '\(0xff)';
		    chan*.ch_newError := CH_BADREAD;
		else
		    unit.ch_file*.f_inMax := len;
		    if len = 0L0 then
			ch := '\e';
		    else
			ch := unit.ch_file*.f_buffer[0];
			unit.ch_file*.f_buffPos := 0L1;
		    fi;
		fi;
	    else
		unit.ch_file*.f_buffPos := pos + 0L1;
		ch := unit.ch_file*.f_buffer[pos];
	    fi;
	incase CH_PROC:
	    ch := unit.ch_getChar()
	incase CH_CHARS:
	    ch := unit.ch_chars*;
	    chan*.ch_.ch_chars := unit.ch_chars + 1;
	esac;
	if ch = '\e' then
	    chan*.ch_eof := true;
	fi;
	ch
    fi
corp;

/*
 * _d_channelUnGetChar - put an unused character back into an input stream.
 */

proc _d_channelUnGetChar(char ch)void:
    register *CHANNEL chan;

    chan := ChannelStack[ChannelStackNext - 1];
    chan*.ch_gotChar := true;
    chan*.ch_buffer := ch;
corp;

/*
 * _d_channelPutChar - output a character to a text channel.
 */

proc _d_channelPutChar(register char ch)void:
    extern
	_d_writeErr()void,
	_d_lineFlush()void;
    CHANNELUNIT unit;
    register ulong pos, wLen;
    ulong offset;
    register *CHANNEL chan;

    chan := ChannelStack[ChannelStackNext - 1];
    if chan*.ch_error ~= CH_OK then
	_d_writeErr();
    elif chan*.ch_newError ~= CH_OK then
	;
    elif chan*.ch_closed then
	chan*.ch_newError := CH_CLOSED;
    else
	unit.ch_chars := chan*.ch_.ch_chars;
	case chan*.ch_mode
	incase CH_STANDARD:
	    if LineOutputPos = LINE_BUFFER_SIZE then
		_d_lineFlush();
	    fi;
	    LineOutputBuffer[LineOutputPos] := ch;
	    LineOutputPos := LineOutputPos + 1;
	    if ch = '\n' then
		_d_lineFlush();
	    fi;
	incase CH_FILE:
	    /* note: f_outMax is not used for a text output file. */
	    pos := unit.ch_file*.f_buffPos;
	    if pos = unit.ch_file*.f_buffLen then
		offset := 0;
		while
		    wLen := Write(unit.ch_file*.f_fd,
				  &unit.ch_file*.f_buffer[offset], pos);
		    pos := pos - wLen;
		    pos ~= 0 and wLen ~= -1
		do
		    offset := offset + wLen;
		od;
		if wLen = -1 then
		    chan*.ch_newError := CH_DISKFULL;
		fi;
		pos := 0L1;
	    else
		pos := pos + 0L1;
	    fi;
	    unit.ch_file*.f_buffer[pos - 0L1] := ch;
	    unit.ch_file*.f_buffPos := pos;
	incase CH_PROC:
	    unit.ch_putChar(ch);
	incase CH_CHARS:
	    unit.ch_chars* := ch;
	    chan*.ch_.ch_chars := unit.ch_chars + 1;
	esac;
    fi;
corp;

/*
 * _d_topChannel - return a pointer to the current I/O channel.
 */

proc _d_topChannel()*CHANNEL:

    ChannelStack[ChannelStackNext - 1]
corp;

/*
 * _d_IO_initialize - automatically invoked initialization routine. Just set
 *		 the channel stack pointer to point to the first slot;
 *		 set up the Standard channels; and set up the console
 *		 input structure.
 */

proc _d_IO_initialize()void:
    extern _d_channelOpen(CHANNEL chan)void;

    ChannelStackNext := 0;
    StandardInputFd := Input();
    StandardOutputFd := Output();
    StandardInput := &StandardInputChannel;
    StandardInput*.ch_mode := CH_STANDARD;
    StandardInput*.ch_input := true;
    StandardInput*.ch_text := true;
    _d_channelOpen(StandardInputChannel);
    StandardOutput := &StandardOutputChannel;
    StandardOutput*.ch_mode := CH_STANDARD;
    StandardOutput*.ch_input := false;
    StandardOutput*.ch_text := true;
    _d_channelOpen(StandardOutputChannel);
    LineInputMax := 0;
    LineInputPos := 0;
    LineOutputPos := 0;
    NormalWriteOpenMode := MODE_NEWFILE;
    NormalReadOpenMode := MODE_OLDFILE;
    PipeWriteOpenMode := MODE_NEWFILE;
    PipeReadOpenMode := MODE_READONLY;
corp;

/*
 * SetStdIn - allow the user to change the current default input channel.
 */

proc SetStdIn(*CHANNEL chan)*CHANNEL:
    extern LineFlush()void;
    *CHANNEL prev;

    LineFlush();
    prev := StandardInput;
    if chan = nil then
	StandardInput := &StandardInputChannel;
    else
	StandardInput := chan;
    fi;
    prev
corp;

/*
 * SetStdOut - allow the user to change the current default output channel.
 */

proc SetStdOut(*CHANNEL chan)*CHANNEL:
    extern LineFlush()void;
    *CHANNEL prev;

    LineFlush();
    prev := StandardOutput;
    if chan = nil then
	StandardOutput := &StandardOutputChannel;
    else
	StandardOutput := chan;
    fi;
    prev
corp;

/*
 * LineFlush - explicitly flush the current standard output channel.
 */

proc LineFlush()void:
    extern _d_lineFlush()void;
    register *CHANNEL chan;
    register *FILE f;
    register ulong pos, offset, wLen;

    chan := StandardOutput;
    case chan*.ch_mode
    incase CH_STANDARD:
	_d_lineFlush();
    incase CH_FILE:
	f := chan*.ch_.ch_file;
	pos := f*.f_buffPos;
	offset := 0;
	while
	    wLen := Write(f*.f_fd, &f*.f_buffer[offset], pos);
	    pos := pos - wLen;
	    pos ~= 0 and wLen ~= -1
	do
	    offset := offset + wLen;
	od;
	if wLen = - 1 then
	    chan*.ch_newError := CH_DISKFULL;
	fi;
	f*.f_buffPos := 0;
    esac;
corp;

/*
 * _d_lineFlush - flush our standard output line buffer.
 */

proc _d_lineFlush()void:
    register ulong pos, wLen, offset;

    pos := LineOutputPos;
    if pos ~= 0 then
	offset := 0;
	while
	    wLen := Write(StandardOutputFd, &LineOutputBuffer[offset],
			  LineOutputPos);
	    pos := pos - wLen;
	    pos ~= 0 and wLen ~= -1
	do
	    offset := offset + wLen;
	od;
	LineOutputPos := 0;
    fi;
corp;

/*
 * _d_printString - write the passed string to the standard output.
 */

proc _d_printString(register *char message)void:
    register *char p;
    register ulong len, wLen;

    p := message;
    while p* ~= '\e' do
	p := p + 1;
    od;
    len := p - message;
    while
	wLen := Write(StandardOutputFd, message, len);
	len := len - wLen;
	len ~= 0 and wLen ~= -1
    do
	message := message + wLen;
    od;
corp;

/*
 * _d_fopen - general file open routine.
 */

proc _d_fopen(*CHANNEL chan; register *FILE f; register *char name;
	      ulong buffLen)bool:
    extern _d_channelOpen(*CHANNEL chan)void;

    _d_channelOpen(chan);
    chan*.ch_mode := CH_FILE;
    chan*.ch_.ch_file := f;
    f*.f_buffPos := 0L0;
    f*.f_inMax := 0L0;
    f*.f_outMax := 0L0;
    f*.f_buffOffset := POS_UNKNOWN;
    f*.f_buffLen := buffLen;
    f*.f_name := name;
    /* note: MODE_OLDFILE == MODE_READONLY, and both seem to let you both
       read and write, and only work if the file already exists. MODE_NEWFILE
       empties a file it opens, so we shouldn't use it. The V 1.3 pipe:
       device only works properly if the write end is MODE_NEWFILE, however,
       so we do it this way in the hopes of keeping most things happy. There
       is also a possible problem with ReOpen, in that it takes an FD opened
       for reading and wants to write with it.

       Redone to give the user a choice. The defaults are:
	    NormalWriteOpenMode = MODE_NEWFILE	    (empties it!!!!)
	    NormalReadOpenMode = MODE_OLDFILE	    (allows read and write)
	    PipeWriteOpenMode = MODE_NEWFILE
	    PipeReadOpenMode = MODE_READONLY
    */
    f*.f_fd :=
	Open(name,
	     if name* = 'p' and (name + sizeof(char))* = 'i' and
		(name + 2 * sizeof(char))* = 'p' and
		(name + 3 * sizeof(char))* = 'e' and
		(name + 4 * sizeof(char))* = ':'
	     then
		if chan*.ch_input then
		    PipeReadOpenMode
		else
		    PipeWriteOpenMode
		fi
	     else
		if chan*.ch_input then
		    NormalReadOpenMode
		else
		    NormalWriteOpenMode
		fi
	     fi);
    if f*.f_fd = 0L0 then
	chan*.ch_error := CH_NONEXIS;
	false
    else
	true
    fi
corp;

proc SetNormalReadOpenMode(ulong newMode)ulong:
    ulong oldMode;

    oldMode := NormalReadOpenMode;
    NormalReadOpenMode := newMode;
    oldMode
corp;

proc SetNormalWriteOpenMode(ulong newMode)ulong:
    ulong oldMode;

    oldMode := NormalWriteOpenMode;
    NormalWriteOpenMode := newMode;
    oldMode
corp;

proc SetPipeReadOpenMode(ulong newMode)ulong:
    ulong oldMode;

    oldMode := PipeReadOpenMode;
    PipeReadOpenMode := newMode;
    oldMode
corp;

proc SetPipeWriteOpenMode(ulong newMode)ulong:
    ulong oldMode;

    oldMode := PipeWriteOpenMode;
    PipeWriteOpenMode := newMode;
    oldMode
corp;
