#drinc:util.g

uint
    FSIZE = 1024+128,
    BSIZE = 10000;

channel input binary Chin;
channel output binary Chout;

file(FSIZE) File;

proc inError()void:
    ushort ec;

    ec := ioerror(Chin);
    if ec ~= CH_OK then
	case ec
	incase CH_EOF:
	    writeln("input: end-of-file");
	incase CH_CLOSED:
	    writeln("input: closed file");
	incase CH_NONEXIS:
	    writeln("input: non-existent");
	incase CH_DISKFULL:
	    writeln("input: disk full");
	incase CH_BADSEEK:
	    writeln("input: bad seek");
	incase CH_MISSING:
	    writeln("input: missing data");
	incase CH_BADCHAR:
	    writeln("input: bad character");
	incase CH_OVERFLOW:
	    writeln("input: overflow");
	default:
	    writeln("input: other error code");
	esac;
	close(Chout);
	exit(1);
    fi;
corp;

proc outError()void:
    ushort ec;

    ec := ioerror(Chout);
    if ec ~= CH_OK then
	case ec
	incase CH_EOF:
	    writeln("output: end-of-file");
	incase CH_CLOSED:
	    writeln("output: closed file");
	incase CH_NONEXIS:
	    writeln("output: non-existent");
	incase CH_DISKFULL:
	    writeln("output: disk full");
	incase CH_BADSEEK:
	    writeln("output: bad seek");
	incase CH_MISSING:
	    writeln("output: missing data");
	incase CH_BADCHAR:
	    writeln("output: bad character");
	incase CH_OVERFLOW:
	    writeln("output: overflow");
	default:
	    writeln("output: other error code");
	esac;
	close(Chout);
	exit(1);
    fi;
corp;

proc main()void:
    [BSIZE] byte buffer;
    [BSIZE] char cbuf @ buffer;
    ulong fd;
    ulong pos, cnt, len, i;
    char ch;
    ushort ec;

    if not open(Chin, File, "test.dat") then
	if not FileCreate("test.dat") then
	    writeln("Can't create file.");
	    exit(1);
	fi;
	if not open(Chin, File, "test.dat") then
	    writeln("Can't open for input.");
	    exit(1);
	fi;
    fi;
    ReOpen(Chin, Chout);
    inError();
    outError();
    while
	write("> ");
	if read(ch) then
	    ch ~= 'q'
	else
	    if ioerror() = CH_EOF then
		false
	    else
		ch := ' ';
		true
	    fi
	fi
    do
	if ch = ' ' then
	    ;
	elif ch = '?' then
	    writeln("Commands are:");
	    writeln();
	    writeln("q - quit");
	    writeln("p - display current position");
	    writeln("s data - sequential write");
	    writeln("w <pos> data - random write");
	    writeln("r <pos> <count> - random read");
	    writeln();
	elif ch ~= 'r' and ch ~= 'w' and ch ~= 's' and ch ~= 'p' then
	    writeln("Bad command, type '?' for help.");
	elif ch = 'p' then
	   writeln("Current position is ", GetOut(Chout));
	elif ch ~= 's' and not read(pos) then
	    writeln("Bad position.");
	    ec := ioerror();
	else
	    if ch = 'r' then
		if not read(cnt) then
		    writeln("Bad count.");
		    ec := ioerror();
		elif cnt > BSIZE then
		    writeln("Count too big.");
		else
		    cbuf[cnt] := '\e';
		    for i from cnt + 1 upto BSIZE - 1 do
			cbuf[i] := 'X';
		    od;
		    if not SeekOut(Chout, pos) then
			writeln("Bad read seek.");
			inError();
		    fi;
		    len := RawRead(Chin, &buffer[0], cnt);
		    if len ~= cnt then
			writeln("Read only ", len, " bytes.");
			inError();
		    fi;
		    writeln(">>", &cbuf[0], "<<");
		fi;
	    else
		if not read(&cbuf[0]) then
		    writeln("Bad read of data to write.");
		    ec := ioerror();
		else
		    if ch = 'w' then
			if not SeekOut(Chout, pos) then
			    writeln("Bad write seek.");
			    outError();
			fi;
		    fi;
		    cnt := CharsLen(&cbuf[0]);
		    len := RawWrite(Chout, &buffer[0], cnt);
		    if len ~= cnt then
			writeln("Wrote only ", len, " bytes.");
			outError();
		    fi;
		fi;
	    fi;
	fi;
	if not readln() then
	    writeln("Bad readln");
	fi;
    od;
    if not close(Chout) then
	writeln("Bad write close.");
	outError();
    fi;
corp;
